#include<stdio.h>
#include<stdlib.h>
#include<math.h>
#include<string.h>

#define pi 3.141592653589793
#define bufsize 1000
#define Nmax 1000000

struct vertex {
  int previous,next,new_contour;
  double x,y,gamma[3];
};

struct edge {
  int v1,v2;
  double x1,y1,x2,y2,n[3];
};

struct vertex* vertices;
struct edge* edges;
int Nvertices,Nedges;
double eps0,Poisson;
char *filename_parameters, *filename_input, *filename_output;

void get_eps2d(double x, double y, int Nvertices, int Nedges, struct vertex* vertices, struct edge* edges,
  double eps0, double Poisson, double* result_eps, double* result_chi);
void set_vertices_and_edges();

void read_parameters(char* filename);
void calc_strain(char* filename_in, char* filename_out);

int main(int argc, char* argv[]) {
  if(argc!=4) fprintf(stderr,"Error!\n  main: wrong number of arguments.\n"),exit(1);
  filename_parameters=argv[1];
  filename_input=argv[2];
  filename_output=argv[3];
  read_parameters(filename_parameters);
  set_vertices_and_edges();
  calc_strain(filename_input,filename_output);
  free(vertices);
  free(edges);
  return 0;
}


// -----------------------


double get_log_d(double x, double y, double x1, double y1);
double get_theta(double x, double y, double x1, double y1, double x2, double y2);
double get_contour_area(int k);
void set_gamma_for_all_vertices();
void set_n_for_all_edges();


void get_eps2d(double x, double y, int Nvertices, int Nedges, struct vertex* vertices, struct edge* edges,
  double eps0, double Poisson, double* result_eps, double* result_chi) {
  int i,k;
  double lambda,epsxx,epsyy,epsxy,theta,log_d,chi;
  lambda=eps0*(1+Poisson)/(4*pi*(1-Poisson));
  epsxx=epsyy=epsxy=chi=0;
  for (i=1; i<=Nedges; i++) {
    theta=get_theta(x,y,edges[i].x1,edges[i].y1,edges[i].x2,edges[i].y2);
	epsxx-=2*lambda*edges[i].n[0]*theta;
	epsyy-=2*lambda*edges[i].n[1]*theta;
	epsxy-=2*lambda*edges[i].n[2]*theta;
	chi-=theta/(2*pi);
  }
  if (fabs(chi)<1e-8) chi=0;
  if (fabs(chi-1)<1e-8) chi=1;
  epsxx-=eps0*chi;
  epsyy-=eps0*chi;
  for (k=1; k<=Nvertices; k++) {
    log_d=get_log_d(x,y,vertices[k].x,vertices[k].y);
	epsxx+=2*lambda*vertices[k].gamma[0]*log_d;
	epsyy+=2*lambda*vertices[k].gamma[1]*log_d;
	epsxy+=2*lambda*vertices[k].gamma[2]*log_d;
  }
  result_eps[0]=epsxx;
  result_eps[1]=epsyy;
  result_eps[2]=epsxy;
  *result_chi=chi;
}

double get_log_d(double x, double y, double x1, double y1) {
  double r2=(x-x1)*(x-x1)+(y-y1)*(y-y1);
  if (r2>0) return 0.5*log(r2);
  else return -1000;
}

double get_theta(double x, double y, double x1, double y1, double x2, double y2) {
  double sin_theta=(x-x2)*(y-y1)-(x-x1)*(y-y2), cos_theta=(x-x1)*(x-x2)+(y-y1)*(y-y2);
  if (sin_theta==0 && cos_theta==0) return 0;
  else return atan2(sin_theta,cos_theta);
}


void set_vertices_and_edges() {
  int i,k,k0,k1,k2,prev_,next_;
  // there we assume that the array "vertices" is created, and fields "x", "y", "next", "new_contour" are filled
  // check fields "next", "new_contour"
  for (k=1; k<=Nvertices; k++) {
    if (vertices[k].next<1 || vertices[k].next>Nvertices || vertices[k].next==k)
      fprintf(stderr,"Error!\n  set_vertices_and_edges: internal error.\n"),exit(1);
    if (!(vertices[k].new_contour==0 || vertices[k].new_contour==1))
      fprintf(stderr,"Error!\n  set_vertices_and_edges: internal error.\n"),exit(1);
  }
  // fill the field "previous"
  for (k=1; k<=Nvertices; k++) vertices[k].previous=0;
  for (k=1; k<=Nvertices; k++)
    if (vertices[k].new_contour==1) {
	  k1=k;
	  while (1) {
		k0=k1; k1=vertices[k1].next;
		if (vertices[k1].previous!=0) fprintf(stderr,"Error!\n  set_vertices_and_edges: internal error.\n"),exit(1);
		vertices[k1].previous=k0;
		if (k1==k) break;
		if (vertices[k1].new_contour==1) fprintf(stderr,"Error!\n  set_vertices_and_edges: internal error.\n"),exit(1);
	  }
  }
  // for all contours, provide positive contour areas
  for (k=1; k<=Nvertices; k++)
    if (vertices[k].new_contour==1)
	  if (get_contour_area(k)<0) {
	    k1=k;
		while (1) {
		  next_=vertices[k1].next;
		  prev_=vertices[k1].previous;
		  vertices[k1].next    =prev_;
		  vertices[k1].previous=next_;
		  k1=next_;
		  if (k1==k) break;
		}
	  }
  // create the array "edges", fill fields "v1", "v2", "x1", "x2", "y1", "y2"
  Nedges=Nvertices;
  edges=(struct edge*) malloc((Nedges+1)*sizeof(struct edge));
  if (edges==NULL) fprintf(stderr,"Error!\n  set_vertices_and_edges: can't allocate memory.\n"),exit(1);
  for (k=1; k<=Nvertices; k++) {
    i=k; k2=vertices[k].next;
	edges[i].v1=k;
	edges[i].v2=k2;
	edges[i].x1=vertices[k].x;
	edges[i].y1=vertices[k].y;
	edges[i].x2=vertices[k2].x;
	edges[i].y2=vertices[k2].y;
  }
  // fill fields "gamma", "n"
  set_gamma_for_all_vertices();
  set_n_for_all_edges();
}


double get_contour_area(int k) {
  int k1,c;
  double x1,y1,x2,y2,S;
  k1=k; S=0;
  for (c=0; c<=Nvertices; c++) {
    if (c==Nvertices) fprintf(stderr,"Error!\n  get_contour_area: internal error.\n"),exit(1);
	x1=vertices[k1].x;
	y1=vertices[k1].y;
	k1=vertices[k1].next;
	x2=vertices[k1].x;
	y2=vertices[k1].y;
	S+=0.5*(x1-x2)*(y1+y2);
	if (k1==k) break;
  }
  return S;
}

void set_gamma_for_all_vertices() {
  int k,k1,k2;
  double x0,y0,x1,y1,x2,y2,n1x,n1y,b1x,b1y,n2x,n2y,b2x,b2y,l;
  for (k=1; k<=Nvertices; k++) {
    k1=vertices[k].previous;
	k2=vertices[k].next;
	x0=vertices[k].x;
	y0=vertices[k].y;
	x1=vertices[k1].x;
	y1=vertices[k1].y;
	x2=vertices[k2].x;
	y2=vertices[k2].y;
    n1x=y0-y1; 	n1y=x1-x0;
    n2x=y2-y0; 	n2y=x0-x2;
	l=sqrt(n1x*n1x+n1y*n1y);
	if (l==0) fprintf(stderr,"Error!\n  set_gamma_for_all_vertices: two adjacent vertices coinside.\n"),exit(1);
	n1x=n1x/l; n1y=n1y/l;
	l=sqrt(n2x*n2x+n2y*n2y);
	if (l==0) fprintf(stderr,"Error!\n  set_gamma_for_all_vertices: two adjacent vertices coinside.\n"),exit(1);
	n2x=n2x/l; n2y=n2y/l;
	b1x=-n1y; b1y=n1x;
	b2x=n2y; b2y=-n2x;
	vertices[k].gamma[0]=n1x*b1x+n2x*b2x;
	vertices[k].gamma[1]=n1y*b1y+n2y*b2y;
	vertices[k].gamma[2]=n1x*b1y+n2x*b2y;
  }
}

void set_n_for_all_edges() {
  int i;
  double nx,ny,l;
  for (i=1; i<=Nedges; i++) {
    nx=edges[i].y2-edges[i].y1;
	ny=edges[i].x1-edges[i].x2;
	l=sqrt(nx*nx+ny*ny);
	if (l==0) fprintf(stderr,"Error!\n  set_n_for_all_edges: two adjacent vertices coinside.\n"),exit(1);
	nx=nx/l; ny=ny/l;
	edges[i].n[0]=nx*nx;
	edges[i].n[1]=ny*ny;
	edges[i].n[2]=nx*ny;
  }
}


// ------------------------


int alex_fgets (char* s, int n, FILE* in);

void calc_strain(char* filename_in, char* filename_out) {
  double eps[3],chi,x,y;
  FILE *in, *out;
  char buf[bufsize];
  int f,n;
  in=fopen(filename_in,"r");
  if (in==NULL) fprintf(stderr,"Error!\n  calc_strain: can't open the file %s for reading.\n",filename_in),exit(1);
  out=fopen(filename_out,"w");
  if (out==NULL) fprintf(stderr,"Error!\n  calc_strain: can't open the file %s for writing.\n",filename_out),exit(1);
  while (1) {
    f=alex_fgets(buf,bufsize,in);
	if (f==0) break;
	n=sscanf(buf,"%lf%lf",&x,&y);
	if (n!=2) fprintf(out,"%s\n",buf);
	else {
      get_eps2d(x,y,Nvertices,Nedges,vertices,edges,eps0,Poisson,eps,&chi);
      fprintf(out,"%s   %.2lf  %10lf  %10lf  %10lf\n",buf,chi,eps[0],eps[1],eps[2]);
	}
  }
  fclose(in);
  fclose(out);
}

int alex_fgets (char* s, int n, FILE* in) {
  int i,f;
  char c;
  if (n<2) return 0;
  i=0; f=0;
  while (1) {
    if (i>=n-1) break;
	c=fgetc(in);
	if (c==EOF) break;
	if (c=='\r') continue;
	f=1;
	if (c=='\n') break;
	s[i]=c;
	i++;
  }
  s[i]=0;
  return f;
}


// ------------------------


void read_parameters(char* filename) {
  int vertices_done=0, eps0_done=0, Poisson_done=0;
  int k,new_contour_f,last_contour;
  int n,code,save_code;
  double dd;
  FILE* in;
  char s[bufsize+1], format_string[30];
  sprintf(format_string,"%%%ds",bufsize-1);
  in=fopen(filename,"r");
  if (in==NULL) fprintf(stderr,"Error!\n  read_parameters: can't open the file %s for reading.\n",filename),exit(1);
  code=0; save_code=-1;
  while (code>=0) {
    n=fscanf(in,format_string,s);
	if (n<1 || s[0]==0) code=-2;
	if (!strcmp(s,"_[") || !strcmp(s,"_]")) {
	  if (!strcmp(s,"_[") && code!=1)
	    save_code=code, code=1;
	  if (!strcmp(s,"_]") && code==1)
	    code=save_code, save_code=-1;
	}
	else
	switch (code) {
	case 0:  if (!strcmp(s,"_vertices")) code=11;
              if (!strcmp(s,"_eps0"))     code=21;
	          if (!strcmp(s,"_Poisson"))  code=31;
	          if (!strcmp(s,"_end"))      code=-1;
			  if (s[0]=='_' && code==0)  code=-10;
	          break;
	case 1:  break;
	case 11: if (vertices_done) {code=-3; break;}
              vertices_done=1;
              n=sscanf(s,"%d",&Nvertices);
              if (n<1) {code=-4; break;}
              if (Nvertices<3 || Nvertices>Nmax) {code=-5; break;}
              vertices=(struct vertex*) malloc((Nvertices+1)*sizeof(struct vertex));
              if (vertices==NULL) fprintf(stderr,"Error!\n  read_parameters: can't allocate memory.\n"),exit(1);
              for (k=1; k<Nvertices; k++) vertices[k].next=k+1;
              vertices[Nvertices].next=1;
              k=1; new_contour_f=1; last_contour=1;
              code=12;
              break;
    case 12: if (!strcmp(s,"@")) {code=13; break;}
              if (!strcmp(s,"@@")) {new_contour_f=1; if(k>1) {vertices[k-1].next=last_contour; if (last_contour>=k-2) code=-11;} last_contour=k; break;}
              if (s[0]=='_' && k<Nvertices+1) {code=-6; break;}
              if (s[0]=='_' && k>Nvertices+1) {code=-7; break;}
              if (s[0]=='_' && last_contour<=Nvertices) {vertices[k-1].next=last_contour; if (last_contour>=k-2) code=-11;}
              if (!strcmp(s,"_vertices")) code=11;
              if (!strcmp(s,"_eps0"))     code=21;
	          if (!strcmp(s,"_Poisson"))  code=31;
	          if (!strcmp(s,"_end"))      code=-1;
			  if (s[0]=='_' && code==12) code=-10;
	          break;
    case 13: if (k>Nvertices) {code=-7; break;}
              vertices[k].new_contour=new_contour_f;
              new_contour_f=0;
              n=sscanf(s,"%lf",&dd);
              if (n<1) {code=-4; break;}
              vertices[k].x=dd;
              code=14;
              break;
    case 14: n=sscanf(s,"%lf",&dd);
              if (n<1) {code=-4; break;}
              vertices[k].y=dd;
              k++;
              code=12;
              break;
    case 21: eps0_done=1;
              n=sscanf(s,"%lf",&eps0);
              if (n<1) {code=-8; break;}
              code=0;
              break;
    case 31: Poisson_done=1;
              n=sscanf(s,"%lf",&Poisson);
              if (n<1) {code=-9; break;}
              code=0;
              break;
	}
	switch (code) {
	case -2:  fprintf(stderr,"Error!\n  read_parameters: keyword '_end' not found.\n"),exit(1);
	case -3:  fprintf(stderr,"Error!\n  read_parameters: section '_vertices' appears twice.\n"),exit(1);
	case -4:  fprintf(stderr,"Error!\n  read_parameters: error in section '_vertices'.\n"),exit(1);
	case -5:  fprintf(stderr,"Error!\n  read_parameters: number of vertices is out of range.\n"),exit(1);
	case -6:  fprintf(stderr,"Error!\n  read_parameters: too few vertices.\n"),exit(1);
	case -7:  fprintf(stderr,"Error!\n  read_parameters: too many vertices.\n"),exit(1);
	case -8:  fprintf(stderr,"Error!\n  read_parameters: error in section '_eps0'.\n"),exit(1);
	case -9:  fprintf(stderr,"Error!\n  read_parameters: error in section '_Poisson'.\n"),exit(1);
	case -10: fprintf(stderr,"Error!\n  read_parameters: unknown keyword '%s'.\n",s),exit(1);
	case -11: fprintf(stderr,"Error!\n  read_parameters: too few vertices in a contour.\n",s),exit(1);
	}
  }
  fclose(in);
  if (code<-1) fprintf(stderr,"Error!\n  read_parameters: internal error.\n"),exit(1);
  if (!vertices_done) fprintf(stderr,"Error!\n  read_parameters: keyword '_vertices' not found.\n"),exit(1);
  if (!eps0_done)     fprintf(stderr,"Error!\n  read_parameters: keyword '_eps0' not found.\n")    ,exit(1);
  if (!Poisson_done)  fprintf(stderr,"Error!\n  read_parameters: keyword '_Poisson' not found.\n") ,exit(1);
  // output for debugging
  /*
  printf("eps0=%lf   Poisson=%lf   Nvertices=%d\n",eps0,Poisson,Nvertices);
  for (k=1; k<=Nvertices; k++)
    printf("vertex %d:  x=%lf  y=%lf  next=%d  new_contour=%d\n",k,vertices[k].x,vertices[k].y,vertices[k].next,vertices[k].new_contour);
  */
}


